# Copyright (c) 2012 Nicira, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at:
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
vswitch module allows its callers to interact with OVS DB.
"""
from . import util


def ovs_vsctl_add_bridge(bridge):
    """
    This function creates an OVS bridge.
    """
    ret, _out, _err = util.start_process(["ovs-vsctl", "add-br", bridge])
    return ret


def ovs_vsctl_del_bridge(bridge):
    """
    This function deletes the OVS bridge.
    """
    ret, _out, _err = util.start_process(["ovs-vsctl", "del-br", bridge])
    return ret


def ovs_vsctl_del_pbridge(bridge, iface):
    """
    This function deletes the OVS bridge and assigns the bridge IP address
    back to the iface.
    """
    (ip_addr, mask) = util.interface_get_ip(bridge)
    util.interface_assign_ip(iface, ip_addr, mask)
    util.interface_up(iface)
    util.move_routes(bridge, iface)
    return ovs_vsctl_del_bridge(bridge)


def ovs_vsctl_is_ovs_bridge(bridge):
    """
    This function verifies whether given port is an OVS bridge. If it is an
    OVS bridge then it will return True.
    """
    ret, _out, _err = util.start_process(["ovs-vsctl", "br-exists", bridge])
    return ret == 0


def ovs_vsctl_add_port_to_bridge(bridge, iface):
    """
    This function adds given interface to the bridge.
    """
    ret, _out, _err = util.start_process(["ovs-vsctl", "add-port", bridge,
                                          iface])
    return ret


def ovs_vsctl_del_port_from_bridge(port):
    """
    This function removes given port from a OVS bridge.
    """
    ret, _out, _err = util.start_process(["ovs-vsctl", "del-port", port])
    return ret


def ovs_vsctl_set(table, record, column, key, value):
    """
    This function allows to alter the OVS database. If column is a map, then
    caller should also set the key, otherwise the key should be left as an
    empty string.
    """
    if key is None:
        index = column
    else:
        index = "%s:%s" % (column, key)
    index_value = "%s=%s" % (index, value)
    ret, _out, _err = util.start_process(["ovs-vsctl", "set", table, record,
                                          index_value])
    return ret


def ovs_get_physical_interface(bridge):
    """
    This function tries to figure out which is the physical interface that
    belongs to the bridge. If there are multiple physical interfaces assigned
    to this bridge then it will return the first match.
    """
    ret, out, _err = util.start_process(["ovs-vsctl", "list-ifaces", bridge])

    if ret == 0:
        ifaces = out.splitlines()
        for iface in ifaces:
            ret, out, _err = util.start_process(["ovs-vsctl", "get",
                                                 "Interface", iface, "type"])
            if ret == 0:
                if ('""' in out) or ('system' in out):
                    return iface  # this should be the physical interface
    return None



#!/usr/bin/env python3
#
# Copyright (c) 2022,2023 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at:
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#
# Script information:
# -------------------
# This script allows a developer to quickly identify if the issue at hand
# might be related to the kernel running out of resources or if it really is
# an Open vSwitch issue.
#
# For documentation see the kernel_delay.rst file.
#
#
# Dependencies:
# -------------
#  You need to install the BCC package for your specific platform or build it
#  yourself using the following instructions:
#    https://raw.githubusercontent.com/iovisor/bcc/master/INSTALL.md
#
#  Python needs the following additional packages installed:
#    - pytz
#    - psutil
#
#  You can either install your distribution specific package or use pip:
#    pip install pytz psutil
#
import argparse
import datetime
import os
import pytz
import psutil
import re
import sys
import time

import ctypes as ct

try:
    from bcc import BPF, USDT, USDTException
    from bcc.syscall import syscalls, syscall_name
except ModuleNotFoundError:
    print("ERROR: Can't find the BPF Compiler Collection (BCC) tools!")
    sys.exit(os.EX_OSFILE)

from enum import IntEnum


#
# Actual eBPF source code
#
EBPF_SOURCE = """
#include <linux/irq.h>
#include <linux/sched.h>

#define MONITOR_PID <MONITOR_PID>

enum {
<EVENT_ENUM>
};

struct event_t {
    u64 ts;
    u32 tid;
    u32 id;

    int user_stack_id;
    int kernel_stack_id;

    u32 syscall;
    u64 entry_ts;

};

BPF_RINGBUF_OUTPUT(events, <BUFFER_PAGE_CNT>);
BPF_STACK_TRACE(stack_traces, <STACK_TRACE_SIZE>);
BPF_TABLE("percpu_array", uint32_t, uint64_t, dropcnt, 1);
BPF_TABLE("percpu_array", uint32_t, uint64_t, trigger_miss, 1);

BPF_ARRAY(capture_on, u64, 1);
static inline bool capture_enabled(u64 pid_tgid) {
    int key = 0;
    u64 *ret;

    if ((pid_tgid >> 32) != MONITOR_PID)
        return false;

    ret = capture_on.lookup(&key);
    return ret && *ret == 1;
}

static inline bool capture_enabled__() {
    int key = 0;
    u64 *ret;

    ret = capture_on.lookup(&key);
    return ret && *ret == 1;
}

static struct event_t *get_event(uint32_t id) {
    struct event_t *event = events.ringbuf_reserve(sizeof(struct event_t));

    if (!event) {
        dropcnt.increment(0);
        return NULL;
    }

    event->id = id;
    event->ts = bpf_ktime_get_ns();
    event->tid = bpf_get_current_pid_tgid();

    return event;
}

static int start_trigger() {
    int key = 0;
    u64 *val = capture_on.lookup(&key);

    /* If the value is -1 we can't start as we are still processing the
     * results in userspace. */
    if (!val || *val != 0) {
        trigger_miss.increment(0);
        return 0;
    }

    struct event_t *event = get_event(EVENT_START_TRIGGER);
    if (event) {
       events.ringbuf_submit(event, 0);
       *val = 1;
    } else {
        trigger_miss.increment(0);
    }
    return 0;
}

static int stop_trigger() {
    int key = 0;
    u64 *val = capture_on.lookup(&key);

    if (!val || *val != 1)
        return 0;

    struct event_t *event = get_event(EVENT_STOP_TRIGGER);

    if (event)
       events.ringbuf_submit(event, 0);

    if (val)
        *val = -1;

    return 0;
}

<START_TRIGGER>
<STOP_TRIGGER>


/*
 * For the syscall monitor the following probes get installed.
 */
struct syscall_data_t {
    u64 count;
    u64 total_ns;
    u64 worst_ns;
};

struct syscall_data_key_t {
    u32 pid;
    u32 tid;
    u32 syscall;
};

BPF_HASH(syscall_start, u64, u64);
BPF_HASH(syscall_data, struct syscall_data_key_t, struct syscall_data_t);

TRACEPOINT_PROBE(raw_syscalls, sys_enter) {
    u64 pid_tgid = bpf_get_current_pid_tgid();

    if (!capture_enabled(pid_tgid))
       return 0;

    u64 t = bpf_ktime_get_ns();
    syscall_start.update(&pid_tgid, &t);

    return 0;
}

TRACEPOINT_PROBE(raw_syscalls, sys_exit) {
    struct syscall_data_t *val, zero = {};
    struct syscall_data_key_t key;

    u64 pid_tgid = bpf_get_current_pid_tgid();

    if (!capture_enabled(pid_tgid))
       return 0;

    key.pid = pid_tgid >> 32;
    key.tid = (u32)pid_tgid;
    key.syscall = args->id;

    u64 *start_ns = syscall_start.lookup(&pid_tgid);

    if (!start_ns)
        return 0;

    val = syscall_data.lookup_or_try_init(&key, &zero);
    if (val) {
        u64 delta = bpf_ktime_get_ns() - *start_ns;
        val->count++;
        val->total_ns += delta;
        if (val->worst_ns == 0 || delta > val->worst_ns)
            val->worst_ns = delta;

        if (<SYSCALL_TRACE_EVENTS>) {
            struct event_t *event = get_event(EVENT_SYSCALL);
            if (event) {
                event->syscall = args->id;
                event->entry_ts = *start_ns;
                if (<STACK_TRACE_ENABLED>) {
                    event->user_stack_id = stack_traces.get_stackid(
                        args, BPF_F_USER_STACK);
                    event->kernel_stack_id = stack_traces.get_stackid(
                        args, 0);
                }
                events.ringbuf_submit(event, 0);
            }
        }
    }
    return 0;
}


/*
 * For measuring the thread run time, we need the following.
 */
struct run_time_data_t {
    u64 count;
    u64 total_ns;
    u64 max_ns;
    u64 min_ns;
};

struct pid_tid_key_t {
    u32  pid;
    u32  tid;
};

BPF_HASH(run_start, u64, u64);
BPF_HASH(run_data, struct pid_tid_key_t, struct run_time_data_t);

static inline void thread_start_run(u64 pid_tgid, u64 ktime)
{
    run_start.update(&pid_tgid, &ktime);
}

static inline void thread_stop_run(u32 pid, u32 tgid, u64 ktime)
{
    u64 pid_tgid = (u64) tgid << 32 | pid;
    u64 *start_ns = run_start.lookup(&pid_tgid);

    if (!start_ns || *start_ns == 0)
        return;

    struct run_time_data_t *val, zero = {};
    struct pid_tid_key_t key = { .pid = tgid,
                                 .tid = pid };

    val = run_data.lookup_or_try_init(&key, &zero);
    if (val) {
        u64 delta = ktime - *start_ns;
        val->count++;
        val->total_ns += delta;
        if (val->max_ns == 0 || delta > val->max_ns)
            val->max_ns = delta;
        if (val->min_ns == 0 || delta < val->min_ns)
            val->min_ns = delta;
    }
    *start_ns = 0;
}


/*
 * For measuring the thread-ready delay, we need the following.
 */
struct ready_data_t {
    u64 count;
    u64 total_ns;
    u64 worst_ns;
};

BPF_HASH(ready_start, u64, u64);
BPF_HASH(ready_data, struct pid_tid_key_t, struct ready_data_t);

static inline int sched_wakeup__(u32 pid, u32 tgid)
{
    u64 pid_tgid = (u64) tgid << 32 | pid;

    if (!capture_enabled(pid_tgid))
        return 0;

    u64 t = bpf_ktime_get_ns();
    ready_start.update(&pid_tgid, &t);
    return 0;
}

RAW_TRACEPOINT_PROBE(sched_wakeup)
{
    struct task_struct *t = (struct task_struct *)ctx->args[0];
    return sched_wakeup__(t->pid, t->tgid);
}

RAW_TRACEPOINT_PROBE(sched_wakeup_new)
{
    struct task_struct *t = (struct task_struct *)ctx->args[0];
    return sched_wakeup__(t->pid, t->tgid);
}

RAW_TRACEPOINT_PROBE(sched_switch)
{
    struct task_struct *prev = (struct task_struct *)ctx->args[1];
    struct task_struct *next= (struct task_struct *)ctx->args[2];
    u64 ktime = 0;

    if (!capture_enabled__())
        return 0;

    if (prev-><STATE_FIELD> == TASK_RUNNING && prev->tgid == MONITOR_PID)
        sched_wakeup__(prev->pid, prev->tgid);

    if (prev->tgid == MONITOR_PID) {
        ktime = bpf_ktime_get_ns();
        thread_stop_run(prev->pid, prev->tgid, ktime);
    }

    u64 pid_tgid = (u64)next->tgid << 32 | next->pid;

    if (next->tgid != MONITOR_PID)
        return 0;

    if (ktime == 0)
        ktime = bpf_ktime_get_ns();

    u64 *start_ns = ready_start.lookup(&pid_tgid);

    if (start_ns && *start_ns != 0) {

        struct ready_data_t *val, zero = {};
        struct pid_tid_key_t key = { .pid = next->tgid,
                                     .tid = next->pid };

        val = ready_data.lookup_or_try_init(&key, &zero);
        if (val) {
            u64 delta = ktime - *start_ns;
            val->count++;
            val->total_ns += delta;
            if (val->worst_ns == 0 || delta > val->worst_ns)
                val->worst_ns = delta;
        }
        *start_ns = 0;
    }

    thread_start_run(pid_tgid, ktime);
    return 0;
}


/*
 * For measuring the hard irq time, we need the following.
 */
struct hardirq_start_data_t {
    u64  start_ns;
    char irq_name[32];
};

struct hardirq_data_t {
    u64 count;
    u64 total_ns;
    u64 worst_ns;
};

struct hardirq_data_key_t {
    u32 pid;
    u32 tid;
    char irq_name[32];
};

BPF_HASH(hardirq_start, u64, struct hardirq_start_data_t);
BPF_HASH(hardirq_data, struct hardirq_data_key_t, struct hardirq_data_t);

TRACEPOINT_PROBE(irq, irq_handler_entry)
{
    u64 pid_tgid = bpf_get_current_pid_tgid();

    if (!capture_enabled(pid_tgid))
        return 0;

    struct hardirq_start_data_t data = {};

    data.start_ns = bpf_ktime_get_ns();
    TP_DATA_LOC_READ_STR(&data.irq_name, name, sizeof(data.irq_name));
    hardirq_start.update(&pid_tgid, &data);
    return 0;
}

TRACEPOINT_PROBE(irq, irq_handler_exit)
{
    u64 pid_tgid = bpf_get_current_pid_tgid();

    if (!capture_enabled(pid_tgid))
        return 0;

    struct hardirq_start_data_t *data;
    data = hardirq_start.lookup(&pid_tgid);
    if (!data || data->start_ns == 0)
        return 0;

    if (args->ret != IRQ_NONE) {
        struct hardirq_data_t *val, zero = {};
        struct hardirq_data_key_t key = { .pid = pid_tgid >> 32,
                                          .tid = (u32)pid_tgid };

        bpf_probe_read_kernel(&key.irq_name, sizeof(key.irq_name),
                              data->irq_name);
        val = hardirq_data.lookup_or_try_init(&key, &zero);
        if (val) {
            u64 delta = bpf_ktime_get_ns() - data->start_ns;
            val->count++;
            val->total_ns += delta;
            if (val->worst_ns == 0 || delta > val->worst_ns)
                val->worst_ns = delta;
        }
    }

    data->start_ns = 0;
    return 0;
}


/*
 * For measuring the soft irq time, we need the following.
 */
struct softirq_start_data_t {
    u64 start_ns;
    u32 vec_nr;
};

struct softirq_data_t {
    u64 count;
    u64 total_ns;
    u64 worst_ns;
};

struct softirq_data_key_t {
    u32 pid;
    u32 tid;
    u32 vec_nr;
};

BPF_HASH(softirq_start, u64, struct softirq_start_data_t);
BPF_HASH(softirq_data, struct softirq_data_key_t, struct softirq_data_t);

TRACEPOINT_PROBE(irq, softirq_entry)
{
    u64 pid_tgid = bpf_get_current_pid_tgid();

    if (!capture_enabled(pid_tgid))
        return 0;

    struct softirq_start_data_t data = {};

    data.start_ns = bpf_ktime_get_ns();
    data.vec_nr = args->vec;
    softirq_start.update(&pid_tgid, &data);
    return 0;
}

TRACEPOINT_PROBE(irq, softirq_exit)
{
    u64 pid_tgid = bpf_get_current_pid_tgid();

    if (!capture_enabled(pid_tgid))
        return 0;

    struct softirq_start_data_t *data;
    data = softirq_start.lookup(&pid_tgid);
    if (!data || data->start_ns == 0)
        return 0;

    struct softirq_data_t *val, zero = {};
    struct softirq_data_key_t key = { .pid = pid_tgid >> 32,
                                      .tid = (u32)pid_tgid,
                                      .vec_nr = data->vec_nr};

    val = softirq_data.lookup_or_try_init(&key, &zero);
    if (val) {
        u64 delta = bpf_ktime_get_ns() - data->start_ns;
        val->count++;
        val->total_ns += delta;
        if (val->worst_ns == 0 || delta > val->worst_ns)
            val->worst_ns = delta;
    }

    data->start_ns = 0;
    return 0;
}
"""


#
# time_ns()
#
try:
    from time import time_ns
except ImportError:
    # For compatibility with Python <= v3.6.
    def time_ns():
        now = datetime.datetime.now()
        return int(now.timestamp() * 1e9)


#
# Probe class to use for the start/stop triggers
#
class Probe(object):
    '''
    The goal for this object is to support as many as possible
    probe/events as supported by BCC. See
       https://github.com/iovisor/bcc/blob/master/docs/reference_guide.md#events--arguments
    '''
    def __init__(self, probe, pid=None):
        self.pid = pid
        self.text_probe = probe
        self._parse_text_probe()

    def __str__(self):
        if self.probe_type == "usdt":
            return "[{}]; {}:{}:{}".format(self.text_probe, self.probe_type,
                                           self.usdt_provider, self.usdt_probe)
        elif self.probe_type == "trace":
            return "[{}]; {}:{}:{}".format(self.text_probe, self.probe_type,
                                           self.trace_system, self.trace_event)
        elif self.probe_type == "kprobe" or self.probe_type == "kretprobe":
            return "[{}]; {}:{}".format(self.text_probe, self.probe_type,
                                        self.kprobe_function)
        elif self.probe_type == "uprobe" or self.probe_type == "uretprobe":
            return "[{}]; {}:{}".format(self.text_probe, self.probe_type,
                                        self.uprobe_function)
        else:
            return "[{}] <{}:unknown probe>".format(self.text_probe,
                                                    self.probe_type)

    def _raise(self, error):
        raise ValueError("[{}]; {}".format(self.text_probe, error))

    def _verify_kprobe_probe(self):
        # Nothing to verify for now, just return.
        return

    def _verify_trace_probe(self):
        # Nothing to verify for now, just return.
        return

    def _verify_uprobe_probe(self):
        # Nothing to verify for now, just return.
        return

    def _verify_usdt_probe(self):
        if not self.pid:
            self._raise("USDT probes need a valid PID.")

        usdt = USDT(pid=self.pid)

        for probe in usdt.enumerate_probes():
            if probe.provider.decode("utf-8") == self.usdt_provider and \
               probe.name.decode("utf-8") == self.usdt_probe:
                return

        self._raise("Can't find UDST probe '{}:{}'".format(self.usdt_provider,
                                                           self.usdt_probe))

    def _parse_text_probe(self):
        '''
        The text probe format is defined as follows:
          <probe_type>:<probe_specific>

        Types:
          USDT:      u|usdt:<provider>:<probe>
          TRACE:     t|trace:<system>:<event>
          KPROBE:    k|kprobe:<kernel_function>
          KRETPROBE: kr|kretprobe:<kernel_function>
          UPROBE:    up|uprobe:<function>
          URETPROBE: ur|uretprobe:<function>
        '''
        args = self.text_probe.split(":")
        if len(args) <= 1:
            self._raise("Can't extract probe type.")

        if args[0] not in ["k", "kprobe", "kr", "kretprobe", "t", "trace",
                           "u", "usdt", "up", "uprobe", "ur", "uretprobe"]:
            self._raise("Invalid probe type '{}'".format(args[0]))

        self.probe_type = "kprobe" if args[0] == "k" else args[0]
        self.probe_type = "kretprobe" if args[0] == "kr" else self.probe_type
        self.probe_type = "trace" if args[0] == "t" else self.probe_type
        self.probe_type = "usdt" if args[0] == "u" else self.probe_type
        self.probe_type = "uprobe" if args[0] == "up" else self.probe_type
        self.probe_type = "uretprobe" if args[0] == "ur" else self.probe_type

        if self.probe_type == "usdt":
            if len(args) != 3:
                self._raise("Invalid number of arguments for USDT")

            self.usdt_provider = args[1]
            self.usdt_probe = args[2]
            self._verify_usdt_probe()

        elif self.probe_type == "trace":
            if len(args) != 3:
                self._raise("Invalid number of arguments for TRACE")

            self.trace_system = args[1]
            self.trace_event = args[2]
            self._verify_trace_probe()

        elif self.probe_type == "kprobe" or self.probe_type == "kretprobe":
            if len(args) != 2:
                self._raise("Invalid number of arguments for K(RET)PROBE")
            self.kprobe_function = args[1]
            self._verify_kprobe_probe()

        elif self.probe_type == "uprobe" or self.probe_type == "uretprobe":
            if len(args) != 2:
                self._raise("Invalid number of arguments for U(RET)PROBE")
            self.uprobe_function = args[1]
            self._verify_uprobe_probe()

    def _get_kprobe_c_code(self, function_name, function_content):
        #
        # The kprobe__* do not require a function name, so it's
        # ignored in the code generation.
        #
        return """
int {}__{}(struct pt_regs *ctx) {{
    {}
}}
""".format(self.probe_type, self.kprobe_function, function_content)

    def _get_trace_c_code(self, function_name, function_content):
        #
        # The TRACEPOINT_PROBE() do not require a function name, so it's
        # ignored in the code generation.
        #
        return """
TRACEPOINT_PROBE({},{}) {{
    {}
}}
""".format(self.trace_system, self.trace_event, function_content)

    def _get_uprobe_c_code(self, function_name, function_content):
        return """
int {}(struct pt_regs *ctx) {{
    {}
}}
""".format(function_name, function_content)

    def _get_usdt_c_code(self, function_name, function_content):
        return """
int {}(struct pt_regs *ctx) {{
    {}
}}
""".format(function_name, function_content)

    def get_c_code(self, function_name, function_content):
        if self.probe_type == "kprobe" or self.probe_type == "kretprobe":
            return self._get_kprobe_c_code(function_name, function_content)
        elif self.probe_type == "trace":
            return self._get_trace_c_code(function_name, function_content)
        elif self.probe_type == "uprobe" or self.probe_type == "uretprobe":
            return self._get_uprobe_c_code(function_name, function_content)
        elif self.probe_type == "usdt":
            return self._get_usdt_c_code(function_name, function_content)

        return ""

    def probe_name(self):
        if self.probe_type == "kprobe" or self.probe_type == "kretprobe":
            return "{}".format(self.kprobe_function)
        elif self.probe_type == "trace":
            return "{}:{}".format(self.trace_system,
                                  self.trace_event)
        elif self.probe_type == "uprobe" or self.probe_type == "uretprobe":
            return "{}".format(self.uprobe_function)
        elif self.probe_type == "usdt":
            return "{}:{}".format(self.usdt_provider,
                                  self.usdt_probe)

        return ""


#
# event_to_dict()
#
def event_to_dict(event):
    return dict([(field, getattr(event, field))
                 for (field, _) in event._fields_
                 if isinstance(getattr(event, field), (int, bytes))])


#
# Event enum
#
Event = IntEnum("Event", ["SYSCALL", "START_TRIGGER", "STOP_TRIGGER"],
                start=0)


#
# process_event()
#
def process_event(ctx, data, size):
    global start_trigger_ts
    global stop_trigger_ts

    event = bpf["events"].event(data)
    if event.id == Event.SYSCALL:
        syscall_events.append({"tid": event.tid,
                               "ts_entry": event.entry_ts,
                               "ts_exit": event.ts,
                               "syscall": event.syscall,
                               "user_stack_id": event.user_stack_id,
                               "kernel_stack_id": event.kernel_stack_id})
    elif event.id == Event.START_TRIGGER:
        #
        # This event would have started the trigger already, so all we need to
        # do is record the start timestamp.
        #
        start_trigger_ts = event.ts

    elif event.id == Event.STOP_TRIGGER:
        #
        # This event would have stopped the trigger already, so all we need to
        # do is record the start timestamp.
        stop_trigger_ts = event.ts


#
# next_power_of_two()
#
def next_power_of_two(val):
    np = 1
    while np < val:
        np *= 2
    return np


#
# unsigned_int()
#
def unsigned_int(value):
    try:
        value = int(value)
    except ValueError:
        raise argparse.ArgumentTypeError("must be an integer")

    if value < 0:
        raise argparse.ArgumentTypeError("must be positive")
    return value


#
# unsigned_nonzero_int()
#
def unsigned_nonzero_int(value):
    value = unsigned_int(value)
    if value == 0:
        raise argparse.ArgumentTypeError("must be nonzero")
    return value


#
# get_thread_name()
#
def get_thread_name(pid, tid):
    try:
        with open(f"/proc/{pid}/task/{tid}/comm", encoding="utf8") as f:
            return f.readline().strip("\n")
    except FileNotFoundError:
        pass

    return f"<unknown:{pid}/{tid}>"


#
# get_vec_nr_name()
#
def get_vec_nr_name(vec_nr):
    known_vec_nr = ["hi", "timer", "net_tx", "net_rx", "block", "irq_poll",
                    "tasklet", "sched", "hrtimer", "rcu"]

    if vec_nr < 0 or vec_nr > len(known_vec_nr):
        return f"<unknown:{vec_nr}>"

    return known_vec_nr[vec_nr]


#
# start/stop/reset capture
#
def start_capture():
    bpf["capture_on"][ct.c_int(0)] = ct.c_int(1)


def stop_capture(force=False):
    if force:
        bpf["capture_on"][ct.c_int(0)] = ct.c_int(0xffff)
    else:
        bpf["capture_on"][ct.c_int(0)] = ct.c_int(0)


def capture_running():
    return bpf["capture_on"][ct.c_int(0)].value == 1


def reset_capture():
    bpf["syscall_start"].clear()
    bpf["syscall_data"].clear()
    bpf["run_start"].clear()
    bpf["run_data"].clear()
    bpf["ready_start"].clear()
    bpf["ready_data"].clear()
    bpf["hardirq_start"].clear()
    bpf["hardirq_data"].clear()
    bpf["softirq_start"].clear()
    bpf["softirq_data"].clear()
    bpf["stack_traces"].clear()


#
# Display timestamp
#
def print_timestamp(msg):
    ltz = datetime.datetime.now()
    utc = ltz.astimezone(pytz.utc)
    time_string = "{} @{} ({} UTC)".format(
        msg, ltz.isoformat(), utc.strftime("%H:%M:%S"))
    print(time_string)


#
# process_results()
#
def process_results(syscall_events=None, trigger_delta=None):
    if trigger_delta:
        print_timestamp("# Triggered sample dump, stop-start delta {:,} ns".
                        format(trigger_delta))
    else:
        print_timestamp("# Sample dump")

    #
    # First get a list of all threads we need to report on.
    #
    threads_syscall = {k.tid for k, _ in bpf["syscall_data"].items()
                       if k.syscall != 0xffffffff}

    threads_run = {k.tid for k, _ in bpf["run_data"].items()
                   if k.pid != 0xffffffff}

    threads_ready = {k.tid for k, _ in bpf["ready_data"].items()
                     if k.pid != 0xffffffff}

    threads_hardirq = {k.tid for k, _ in bpf["hardirq_data"].items()
                       if k.pid != 0xffffffff}

    threads_softirq = {k.tid for k, _ in bpf["softirq_data"].items()
                       if k.pid != 0xffffffff}

    threads = sorted(threads_syscall | threads_run | threads_ready |
                     threads_hardirq | threads_softirq,
                     key=lambda x: get_thread_name(options.pid, x))

    #
    # Print header...
    #
    print("{:10} {:16} {}".format("TID", "THREAD", "<RESOURCE SPECIFIC>"))
    print("{:10} {:16} {}".format("-" * 10, "-" * 16, "-" * 76))
    indent = 28 * " "

    #
    # Print all events/statistics per threads.
    #
    poll_id = [k for k, v in syscalls.items() if v == b"poll"][0]
    for thread in threads:

        if thread != threads[0]:
            print("")

        #
        # SYSCALL_STATISTICS
        #
        print("{:10} {:16} {}\n{}{:20} {:>6}  {:>10}  {:>16}  {:>16}".format(
            thread, get_thread_name(options.pid, thread),
            "[SYSCALL STATISTICS]", indent,
            "NAME", "NUMBER", "COUNT", "TOTAL ns", "MAX ns"))

        total_count = 0
        total_ns = 0
        for k, v in sorted(filter(lambda t: t[0].tid == thread,
                                  bpf["syscall_data"].items()),
                           key=lambda kv: -kv[1].total_ns):

            print("{}{:20.20} {:6}  {:10}  {:16,}  {:16,}".format(
                indent, syscall_name(k.syscall).decode("utf-8"), k.syscall,
                v.count, v.total_ns, v.worst_ns))
            if k.syscall != poll_id:
                total_count += v.count
                total_ns += v.total_ns

        if total_count > 0:
            print("{}{:20.20} {:6}  {:10}  {:16,}".format(
                indent, "TOTAL( - poll):", "", total_count, total_ns))

        #
        # THREAD RUN STATISTICS
        #
        print("\n{:10} {:16} {}\n{}{:10}  {:>16}  {:>16}  {:>16}".format(
            "", "", "[THREAD RUN STATISTICS]", indent,
            "SCHED_CNT", "TOTAL ns", "MIN ns", "MAX ns"))

        for k, v in filter(lambda t: t[0].tid == thread,
                           bpf["run_data"].items()):

            print("{}{:10}  {:16,}  {:16,}  {:16,}".format(
                indent, v.count, v.total_ns, v.min_ns, v.max_ns))

        #
        # THREAD READY STATISTICS
        #
        print("\n{:10} {:16} {}\n{}{:10}  {:>16}  {:>16}".format(
            "", "", "[THREAD READY STATISTICS]", indent,
            "SCHED_CNT", "TOTAL ns", "MAX ns"))

        for k, v in filter(lambda t: t[0].tid == thread,
                           bpf["ready_data"].items()):

            print("{}{:10}  {:16,}  {:16,}".format(
                indent, v.count, v.total_ns, v.worst_ns))

        #
        # HARD IRQ STATISTICS
        #
        total_ns = 0
        total_count = 0
        header_printed = False
        for k, v in sorted(filter(lambda t: t[0].tid == thread,
                                  bpf["hardirq_data"].items()),
                           key=lambda kv: -kv[1].total_ns):

            if not header_printed:
                print("\n{:10} {:16} {}\n{}{:20}  {:>10}  {:>16}  {:>16}".
                      format("", "", "[HARD IRQ STATISTICS]", indent,
                             "NAME", "COUNT", "TOTAL ns", "MAX ns"))
                header_printed = True

            print("{}{:20.20}  {:10}  {:16,}  {:16,}".format(
                indent, k.irq_name.decode("utf-8"),
                v.count, v.total_ns, v.worst_ns))

            total_count += v.count
            total_ns += v.total_ns

        if total_count > 0:
            print("{}{:20.20}  {:10}  {:16,}".format(
                indent, "TOTAL:", total_count, total_ns))

        #
        # SOFT IRQ STATISTICS
        #
        total_ns = 0
        total_count = 0
        header_printed = False
        for k, v in sorted(filter(lambda t: t[0].tid == thread,
                                  bpf["softirq_data"].items()),
                           key=lambda kv: -kv[1].total_ns):

            if not header_printed:
                print("\n{:10} {:16} {}\n"
                      "{}{:20} {:>7}  {:>10}  {:>16}  {:>16}".
                      format("", "", "[SOFT IRQ STATISTICS]", indent,
                             "NAME", "VECT_NR", "COUNT", "TOTAL ns", "MAX ns"))
                header_printed = True

            print("{}{:20.20} {:>7}  {:10}  {:16,}  {:16,}".format(
                indent, get_vec_nr_name(k.vec_nr), k.vec_nr,
                v.count, v.total_ns, v.worst_ns))

            total_count += v.count
            total_ns += v.total_ns

        if total_count > 0:
            print("{}{:20.20} {:7}  {:10}  {:16,}".format(
                indent, "TOTAL:", "", total_count, total_ns))

    #
    # Print events
    #
    lost_stack_traces = 0
    if syscall_events:
        stack_traces = bpf.get_table("stack_traces")

        print("\n\n# SYSCALL EVENTS:"
              "\n{}{:>19} {:>19} {:>10} {:16} {:>10}  {}".format(
                  2 * " ", "ENTRY (ns)", "EXIT (ns)", "TID", "COMM",
                  "DELTA (us)", "SYSCALL"))
        print("{}{:19} {:19} {:10} {:16} {:10}  {}".format(
            2 * " ", "-" * 19, "-" * 19, "-" * 10, "-" * 16,
            "-" * 10, "-" * 16))
        for event in syscall_events:
            print("{}{:19} {:19} {:10} {:16} {:10,}  {}".format(
                " " * 2,
                event["ts_entry"], event["ts_exit"], event["tid"],
                get_thread_name(options.pid, event["tid"]),
                int((event["ts_exit"] - event["ts_entry"]) / 1000),
                syscall_name(event["syscall"]).decode("utf-8")))
            #
            # Not sure where to put this, but I'll add some info on stack
            # traces here... Userspace stack traces are very limited due to
            # the fact that bcc does not support dwarf backtraces. As OVS
            # gets compiled without frame pointers we will not see much.
            # If however, OVS does get built with frame pointers, we should not
            # use the BPF_STACK_TRACE_BUILDID as it does not seem to handle
            # the debug symbols correctly. Also, note that for kernel
            # traces you should not use BPF_STACK_TRACE_BUILDID, so two
            # buffers are needed.
            #
            # Some info on manual dwarf walk support:
            #   https://github.com/iovisor/bcc/issues/3515
            #   https://github.com/iovisor/bcc/pull/4463
            #
            if options.stack_trace_size == 0:
                continue

            if event["kernel_stack_id"] < 0 or event["user_stack_id"] < 0:
                lost_stack_traces += 1

            kernel_stack = stack_traces.walk(event["kernel_stack_id"]) \
                if event["kernel_stack_id"] >= 0 else []
            user_stack = stack_traces.walk(event["user_stack_id"]) \
                if event["user_stack_id"] >= 0 else []

            for addr in kernel_stack:
                print("{}{}".format(
                    " " * 10,
                    bpf.ksym(addr, show_module=True,
                             show_offset=True).decode("utf-8", "replace")))

            for addr in user_stack:
                addr_str = bpf.sym(addr, options.pid, show_module=True,
                                   show_offset=True).decode("utf-8", "replace")

                if addr_str == "[unknown]":
                    addr_str += " 0x{:x}".format(addr)

                print("{}{}".format(" " * 10, addr_str))

    #
    # Print any footer messages.
    #
    if lost_stack_traces > 0:
        print("\n#WARNING: We where not able to display {} stack traces!\n"
              "#         Consider increasing the stack trace size using\n"
              "#         the '--stack-trace-size' option.\n"
              "#         Note that this can also happen due to a stack id\n"
              "#         collision.".format(lost_stack_traces))


#
# main()
#
def main():
    #
    # Don't like these globals, but ctx passing does not seem to work with the
    # existing open_ring_buffer() API :(
    #
    global bpf
    global options
    global syscall_events
    global start_trigger_ts
    global stop_trigger_ts

    start_trigger_ts = 0
    stop_trigger_ts = 0

    #
    # Argument parsing
    #
    parser = argparse.ArgumentParser()

    parser.add_argument("-D", "--debug",
                        help="Enable eBPF debugging",
                        type=int, const=0x3f, default=0, nargs="?")
    parser.add_argument("-p", "--pid", metavar="VSWITCHD_PID",
                        help="ovs-vswitch's PID",
                        type=unsigned_int, default=None)
    parser.add_argument("-s", "--syscall-events", metavar="DURATION_NS",
                        help="Record syscall events that take longer than "
                        "DURATION_NS. Omit the duration value to record all "
                        "syscall events",
                        type=unsigned_int, const=0, default=None, nargs="?")
    parser.add_argument("--buffer-page-count",
                        help="Number of BPF ring buffer pages, default 1024",
                        type=unsigned_int, default=1024, metavar="NUMBER")
    parser.add_argument("--sample-count",
                        help="Number of sample runs, default 1",
                        type=unsigned_nonzero_int, default=1, metavar="RUNS")
    parser.add_argument("--sample-interval",
                        help="Delay between sample runs, default 0",
                        type=float, default=0, metavar="SECONDS")
    parser.add_argument("--sample-time",
                        help="Sample time, default 0.5 seconds",
                        type=float, default=0.5, metavar="SECONDS")
    parser.add_argument("--skip-syscall-poll-events",
                        help="Skip poll() syscalls with --syscall-events",
                        action="store_true")
    parser.add_argument("--stack-trace-size",
                        help="Number of unique stack traces that can be "
                        "recorded, default 4096. 0 to disable",
                        type=unsigned_int, default=4096)
    parser.add_argument("--start-trigger", metavar="TRIGGER",
                        help="Start trigger, see documentation for details",
                        type=str, default=None)
    parser.add_argument("--stop-trigger", metavar="TRIGGER",
                        help="Stop trigger, see documentation for details",
                        type=str, default=None)
    parser.add_argument("--trigger-delta", metavar="DURATION_NS",
                        help="Only report event when the trigger duration > "
                             "DURATION_NS, default 0 (all events)",
                        type=unsigned_int, const=0, default=0, nargs="?")

    options = parser.parse_args()

    #
    # Find the PID of the ovs-vswitchd daemon if not specified.
    #
    if not options.pid:
        for proc in psutil.process_iter():
            if "ovs-vswitchd" in proc.name():
                if options.pid:
                    print("ERROR: Multiple ovs-vswitchd daemons running, "
                          "use the -p option!")
                    sys.exit(os.EX_NOINPUT)

                options.pid = proc.pid

    #
    # Error checking on input parameters.
    #
    if not options.pid:
        print("ERROR: Failed to find ovs-vswitchd's PID!")
        sys.exit(os.EX_UNAVAILABLE)

    options.buffer_page_count = next_power_of_two(options.buffer_page_count)

    #
    # Make sure we are running as root, or else we can not attach the probes.
    #
    if os.geteuid() != 0:
        print("ERROR: We need to run as root to attached probes!")
        sys.exit(os.EX_NOPERM)

    #
    # Setup any of the start stop triggers
    #
    if options.start_trigger is not None:
        try:
            start_trigger = Probe(options.start_trigger, pid=options.pid)
        except ValueError as e:
            print(f"ERROR: Invalid start trigger {str(e)}")
            sys.exit(os.EX_CONFIG)
    else:
        start_trigger = None

    if options.stop_trigger is not None:
        try:
            stop_trigger = Probe(options.stop_trigger, pid=options.pid)
        except ValueError as e:
            print(f"ERROR: Invalid stop trigger {str(e)}")
            sys.exit(os.EX_CONFIG)
    else:
        stop_trigger = None

    #
    # Attach probe to running process.
    #
    source = EBPF_SOURCE.replace("<EVENT_ENUM>", "\n".join(
        ["    EVENT_{} = {},".format(
            event.name, event.value) for event in Event]))
    source = source.replace("<BUFFER_PAGE_CNT>",
                            str(options.buffer_page_count))
    source = source.replace("<MONITOR_PID>", str(options.pid))

    if BPF.kernel_struct_has_field(b"task_struct", b"state") == 1:
        source = source.replace("<STATE_FIELD>", "state")
    else:
        source = source.replace("<STATE_FIELD>", "__state")

    poll_id = [k for k, v in syscalls.items() if v == b"poll"][0]
    if options.syscall_events is None:
        syscall_trace_events = "false"
    elif options.syscall_events == 0:
        if not options.skip_syscall_poll_events:
            syscall_trace_events = "true"
        else:
            syscall_trace_events = f"args->id != {poll_id}"
    else:
        syscall_trace_events = "delta > {}".format(options.syscall_events)
        if options.skip_syscall_poll_events:
            syscall_trace_events += f" && args->id != {poll_id}"

    source = source.replace("<SYSCALL_TRACE_EVENTS>",
                            syscall_trace_events)

    source = source.replace("<STACK_TRACE_SIZE>",
                            str(options.stack_trace_size))

    source = source.replace("<STACK_TRACE_ENABLED>", "true"
                            if options.stack_trace_size > 0 else "false")

    #
    # Handle start/stop probes
    #
    if start_trigger:
        source = source.replace("<START_TRIGGER>",
                                start_trigger.get_c_code(
                                    "start_trigger_probe",
                                    "return start_trigger();"))
    else:
        source = source.replace("<START_TRIGGER>", "")

    if stop_trigger:
        source = source.replace("<STOP_TRIGGER>",
                                stop_trigger.get_c_code(
                                    "stop_trigger_probe",
                                    "return stop_trigger();"))
    else:
        source = source.replace("<STOP_TRIGGER>", "")

    #
    # Setup usdt or other probes that need handling trough the BFP class.
    #
    usdt = USDT(pid=int(options.pid))
    try:
        if start_trigger and start_trigger.probe_type == "usdt":
            usdt.enable_probe(probe=start_trigger.probe_name(),
                              fn_name="start_trigger_probe")
        if stop_trigger and stop_trigger.probe_type == "usdt":
            usdt.enable_probe(probe=stop_trigger.probe_name(),
                              fn_name="stop_trigger_probe")

    except USDTException as e:
        print("ERROR: {}".format(
            (re.sub("^", " " * 7, str(e), flags=re.MULTILINE)).strip().
            replace("--with-dtrace or --enable-dtrace",
                    "--enable-usdt-probes")))
        sys.exit(os.EX_OSERR)

    bpf = BPF(text=source, usdt_contexts=[usdt], debug=options.debug)

    if start_trigger:
        try:
            if start_trigger.probe_type == "uprobe":
                bpf.attach_uprobe(name=f"/proc/{options.pid}/exe",
                                  sym=start_trigger.probe_name(),
                                  fn_name="start_trigger_probe",
                                  pid=options.pid)

            if start_trigger.probe_type == "uretprobe":
                bpf.attach_uretprobe(name=f"/proc/{options.pid}/exe",
                                     sym=start_trigger.probe_name(),
                                     fn_name="start_trigger_probe",
                                     pid=options.pid)
        except Exception as e:
            print("ERROR: Failed attaching uprobe start trigger "
                  f"'{start_trigger.probe_name()}';\n         {str(e)}")
            sys.exit(os.EX_OSERR)

    if stop_trigger:
        try:
            if stop_trigger.probe_type == "uprobe":
                bpf.attach_uprobe(name=f"/proc/{options.pid}/exe",
                                  sym=stop_trigger.probe_name(),
                                  fn_name="stop_trigger_probe",
                                  pid=options.pid)

            if stop_trigger.probe_type == "uretprobe":
                bpf.attach_uretprobe(name=f"/proc/{options.pid}/exe",
                                     sym=stop_trigger.probe_name(),
                                     fn_name="stop_trigger_probe",
                                     pid=options.pid)
        except Exception as e:
            print("ERROR: Failed attaching uprobe stop trigger"
                  f"'{stop_trigger.probe_name()}';\n         {str(e)}")
            sys.exit(os.EX_OSERR)

    #
    # If no triggers are configured use the delay configuration
    #
    bpf["events"].open_ring_buffer(process_event)

    sample_count = 0
    while sample_count < options.sample_count:
        sample_count += 1
        syscall_events = []

        if not options.start_trigger:
            print_timestamp("# Start sampling")
            start_capture()
            stop_time = -1 if options.stop_trigger else \
                time_ns() + options.sample_time * 1000000000
        else:
            # For start triggers the stop time depends on the start trigger
            # time, or depends on the stop trigger if configured.
            stop_time = -1 if options.stop_trigger else 0

        while True:
            keyboard_interrupt = False
            try:
                last_start_ts = start_trigger_ts
                last_stop_ts = stop_trigger_ts

                if stop_time > 0:
                    delay = int((stop_time - time_ns()) / 1000000)
                    if delay <= 0:
                        break
                else:
                    delay = -1

                bpf.ring_buffer_poll(timeout=delay)

                if stop_time <= 0 and last_start_ts != start_trigger_ts:
                    print_timestamp(
                        "# Start sampling (trigger@{})".format(
                            start_trigger_ts))

                    if not options.stop_trigger:
                        stop_time = time_ns() + \
                            options.sample_time * 1000000000

                if last_stop_ts != stop_trigger_ts:
                    break

            except KeyboardInterrupt:
                keyboard_interrupt = True
                break

        if options.stop_trigger and not capture_running():
            print_timestamp("# Stop sampling (trigger@{})".format(
                stop_trigger_ts))
        else:
            print_timestamp("# Stop sampling")

        if stop_trigger_ts != 0 and start_trigger_ts != 0:
            trigger_delta = stop_trigger_ts - start_trigger_ts
        else:
            trigger_delta = None

        if not trigger_delta or trigger_delta >= options.trigger_delta:
            stop_capture(force=True)  # Prevent a new trigger to start.
            process_results(syscall_events=syscall_events,
                            trigger_delta=trigger_delta)
        elif trigger_delta:
            sample_count -= 1
            print_timestamp("# Sample dump skipped, delta {:,} ns".format(
                trigger_delta))

        reset_capture()
        stop_capture()

        if keyboard_interrupt:
            break

        if options.sample_interval > 0:
            time.sleep(options.sample_interval)

    #
    # Report lost events.
    #
    dropcnt = bpf.get_table("dropcnt")
    for k in dropcnt.keys():
        count = dropcnt.sum(k).value
        if k.value == 0 and count > 0:
            print("\n# WARNING: Not all events were captured, {} were "
                  "dropped!\n#          Increase the BPF ring buffer size "
                  "with the --buffer-page-count option.".format(count))

    if options.sample_count > 1:
        trigger_miss = bpf.get_table("trigger_miss")
        for k in trigger_miss.keys():
            count = trigger_miss.sum(k).value
            if k.value == 0 and count > 0:
                print("\n# WARNING: Not all start triggers were successful. "
                      "{} were missed due to\n#          slow userspace "
                      "processing!".format(count))


#
# Start main() as the default entry point...
#
if __name__ == "__main__":
    main()



#!/usr/bin/env python3
#
# Copyright (c) 2022 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at:
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Script information:
# -------------------
# reval_monitor.py uses various user-defined tracepoints to get all the
# revalidator-process related variables and will display them in a (dynamic)
# graph. In addition, it will also dump the data to the console
# in a CSV format. Note that all the graphical output can be disabled.
#
# All the USDT events can be saved to a file and than can be used to
# replay the trace offline and still get the plots.
#
# The script can simple be invoked without any options, and it will try
# to find the running ovs-vswitchd instance:
#
#   # ./reval_monitor.py
#   # Starting trace @2022-09-20T04:07:43.588749 (08:07:43 UTC)
#   ts_start, ts_complete, n_flows, n_reval_flows, avg_n_flows, max_n_flows,
#     flow_limit, dump_duration, poll_wait, actual_wait
#   1741367714251645, 1741367714532545, 0, 0, 0, 10000, 69000, 1, 500, 500.52
#   1741368215056961, 1741368215318223, 0, 0, 0, 10000, 69000, 1, 500, 500.55
#   1741368715865871, 1741368716107089, 0, 0, 0, 10000, 69000, 1, 500, 499.48
#   ^C# Stopping trace @2022-09-20T04:07:49.893827 (08:07:49 UTC)
#
#  IMPORTANT NOTE: This script only works when a single datapath is configured!
#  2nd IMPORTANT NOTE: ovs-vswitchd either needs to be built with debug info
#                      or the debug info package needs to be installed!
#
#  The following are the available options:
#
#    usage: reval_monitor.py [-h] [-c] [--buffer-page-count NUMBER]
#                            [-D [DEBUG]] [-g] [--no-ukey-count]
#                            [-p VSWITCHD_PID] [-P PAHOLE] [-r FILE] [-R]
#                            [-u SECONDS] [-w FILE] [-W FILE]
#
#    options:
#      -h, --help            show this help message and exit
#      -c, --compress-output
#                            Compress output, i.e. only dump changes in
#                            the dataset
#      --buffer-page-count NUMBER
#                            Number of BPF ring buffer pages, default
#                            1024
#      -D [DEBUG], --debug [DEBUG]
#                            Enable eBPF debugging
#      -g, --no-gui          Do not use the gui to display plots
#      --no-ukey-count       No revalidate_ukey() counting
#      -p VSWITCHD_PID, --pid VSWITCHD_PID
#                            ovs-vswitch's PID
#      -P PAHOLE, --pahole PAHOLE
#                            Pahole executable to use, default pahole
#      -r FILE, --read-events FILE
#                            Read events from <FILE> instead of
#                            installing tracepoints
#      -R, --no-realtime-plots
#                            Do not show realtime plot while tracing
#      -u SECONDS, --update-interval SECONDS
#                            Seconds to wait between real time update,
#                            default 1
#      -w FILE, --write-events FILE
#                            Write events to <FILE>
#      -W FILE, --write-charts FILE
#                            Write overall charts to <FILE>.png

#                            [-D [DEBUG]] [-g] [--no-ukey-count]
#                            [-p VSWITCHD_PID] [-r FILE] [-R]
#                            [-u SECONDS] [-w FILE] [-W FILE]
#
# The -g option disabled all GUI output of matplotlib, -R only disables the
# real-time plots. As real-time plots are rather slow, the -u option can be
# used to only update the graph every x seconds, which might speed up the
# processing.
#
# The --no-ukey-count option disables counting of the number of flows actually
# being revalidated against the current OpenFlow ruleset. This will not install
# the specific tracepoint which would be called for each flow being
# revalidated.
#
# What is plotted in the graphs (and dumped in the CSV output)?
# - n_flows:       Number of flows active in the system.
# - n_reval_flows: Number of flows that where revalidated against the OpenFlow
#                  ruleset.
# - dump_duration: Time it took to dump and process all flows.
# - avg_n_flows:   Average number of flows in the system.
# - max_n_flows:   Maximum number of flows in the system.
# - flow_limit:    Dynamic flow limit.
# - poll_wait:     Time requested for the poll wait.
# - actual_wait:   Time it took to be woken up.
#
# Dependencies:
#  This script needs the 'readelf' binary to be available. In addition, it also
#  needs pahole to be installed, and it needs a version that is equal or newer
#  than the following commit on the next branch:
#
#    https://git.kernel.org/pub/scm/devel/pahole/pahole.git/?h=next
#      c55b13b9d785 ("WIP: Remove DW_TAG_atomic_type when encoding BTF")
#
#  To use a locally compiled pahole the --pahole option can be used.
#  For example:
#    # ./reval_monitor.py --pahole ~/pahole/build/pahole -g
#    Starting trace @2022-12-20T14:57:26.077815 (13:57:26 UTC)
#    ts_start, ts_complete, n_flows, n_reval_flows, avg_n_flows, max_n_flows, \
#      flow_limit, dump_duration, poll_wait, actual_wait
#    4202771850983494, 4202771851472838, 0, 0, 0, 0, 10000, 1, 500, 15.06
#    4202771866531996, 4202771867713366, 0, 0, 0, 0, 10000, 1, 500, 4.23
#    4202771871941979, 4202771872749915, 0, 0, 0, 0, 10000, 1, 500, 500.02
#    4202772372770361, 4202772373531820, 0, 0, 0, 0, 10000, 1, 500, 499.96
#    4202772873487942, 4202772874514753, 0, 0, 0, 0, 10000, 1, 500, 500.01
#    4202773374528435, 4202773375695054, 0, 0, 0, 0, 10000, 1, 500, 500.01
#    4202773875701559, 4202773876880763, 0, 0, 0, 0, 10000, 1, 500, 500.04
#    4202774376925058, 4202774377905799, 0, 0, 0, 0, 10000, 1, 500, 500.03
#    ^C# Stopping trace @2022-12-20T14:57:40.391730 (13:57:40 UTC)
#

try:
    from bcc import BPF, USDT, USDTException
except ModuleNotFoundError:
    print("WARNING: Can't find the BPF Compiler Collection (BCC) tools!")
    print("         This is NOT problem if you analyzing previously collected"
          " data.\n")

from collections import namedtuple
from enum import IntEnum
from pathlib import Path

import argparse
import ast
import datetime
import re
import subprocess
import sys

import pytz
import psutil
import matplotlib.pyplot as plt

#
# Actual eBPF source code
#
EBPF_SOURCE = """
#include <linux/sched.h>

<OVS_INCLUDE_DEFINITIONS>

enum {
<EVENT_ENUM>
};

struct event_t {
    u64  ts;
    u32  pid;
    u32  id;
    u64  n_flows;
    u32  avg_n_flows;
    u32  max_n_flows;
    u32  flow_limit;
    u32  dump_duration;
    u32  poll_wait;
};


BPF_RINGBUF_OUTPUT(events, <BUFFER_PAGE_CNT>);
BPF_TABLE("percpu_array", uint32_t, uint64_t, dropcnt, 1);

static struct event_t *get_event(uint32_t id) {
    struct event_t *event = events.ringbuf_reserve(sizeof(struct event_t));

    if (!event) {
        dropcnt.increment(0);
        return NULL;
    }

    event->id = id;
    event->ts = bpf_ktime_get_ns();
    event->pid = bpf_get_current_pid_tgid();

    return event;
}

int probe__start_dump(struct pt_regs *ctx) {
    struct event_t *event = get_event(EVENT_START_DUMP);
    if (!event)
        return 1;

    events.ringbuf_submit(event, 0);
    return 0;
};

int probe__sweep_done(struct pt_regs *ctx) {
    struct udpif udpif;

    bpf_usdt_readarg_p(1, ctx, &udpif, sizeof(udpif));

    struct event_t *event = get_event(EVENT_SWEEP_DONE);
    if (!event)
        return 1;

    event->avg_n_flows = udpif.avg_n_flows;
    event->max_n_flows = udpif.max_n_flows;
    event->flow_limit = udpif.flow_limit;
    event->dump_duration = udpif.dump_duration;

    bpf_usdt_readarg(2, ctx, &event->n_flows);
    bpf_usdt_readarg(3, ctx, &event->poll_wait);

    events.ringbuf_submit(event, 0);
    return 0;
};

int probe__reval_entry(struct pt_regs *ctx) {
    struct event_t *event = get_event(EVENT_REVAL_ENTRY);
    if (!event)
        return 1;

    events.ringbuf_submit(event, 0);
    return 0;
};
"""


#
# event_to_dict()
#
def event_to_dict(event):
    return dict([(field, getattr(event, field))
                 for (field, _) in event._fields_
                 if isinstance(getattr(event, field), (int, bytes))])


#
# print_csv_header()
#
def print_csv_header():
    print("ts_start, ts_complete, n_flows, n_reval_flows, avg_n_flows, "
          "max_n_flows, flow_limit, dump_duration, poll_wait, actual_wait")


#
# Event enum
#
Event = IntEnum("Event", ["START_DUMP",
                          "SWEEP_DONE",
                          "REVAL_ENTRY"], start=0)


#
# process_event()
#
def process_event(ctx, data, size):
    event = b['events'].event(data)
    _process_event(event)


def _process_event(event):
    global graph

    if export_file is not None:
        export_file.write("event = {}\n".format(event_to_dict(event)))

    if event.id == Event.START_DUMP and not state['running']:
        start = state["last_start"]
        done = state["last_done"]
        if done and start:
            actual_wait = (event.ts - done.ts) / 1000000
            csv = "{}, {}, {}, {}, {}, {}, {}, {}, {}, {:.2f}".format(
                start.ts, done.ts, done.n_flows, graph.ukey_count,
                done.avg_n_flows, done.max_n_flows, done.flow_limit,
                done.dump_duration, done.poll_wait, actual_wait)

            if graph.base_time == 0:
                graph = graph._replace(base_time=done.ts)

            graph.time.append((done.ts - graph.base_time) / 1000000000)
            graph.n_flows.append(done.n_flows)
            graph.n_reval_flows.append(graph.ukey_count)
            graph.avg_n_flows.append(done.avg_n_flows)
            graph.max_n_flows.append(done.max_n_flows)
            graph.flow_limit.append(done.flow_limit)
            graph.dump_duration.append(done.dump_duration)
            graph.poll_wait.append(done.poll_wait)
            graph.actual_wait.append(actual_wait)

            if not options.no_gui and not options.no_realtime_plots:
                updated_graph = dynamic_plot_update(
                    graph, refresh=options.update_interval)
                if updated_graph is None:
                    raise KeyboardInterrupt
                graph = updated_graph

            if options.compress_output:
                last_csv = state["last_csv"]
                if not last_csv or \
                   csv.split(",")[2:-1] != last_csv.split(",")[2:-1] or \
                   abs((event.ts - done.ts) / 1000000 - done.poll_wait) > 100:
                    print(csv)
                else:
                    state["last_not_printed_csv"] = csv

                state["last_csv"] = csv
            else:
                print(csv)

        state["last_start"] = event
        state['running'] = True
        graph = graph._replace(ukey_count=0)
    elif event.id == Event.SWEEP_DONE and state['running']:
        state["last_done"] = event
        state['running'] = False
    elif event.id == Event.REVAL_ENTRY and state['running']:
        graph = graph._replace(ukey_count=graph.ukey_count + 1)


#
# run_program()
#
def run_program(command):
    try:
        process = subprocess.run(command,
                                 stdout=subprocess.PIPE,
                                 stderr=subprocess.STDOUT,
                                 encoding='utf8',
                                 check=True)

    except subprocess.CalledProcessError as perror:
        return perror.returncode, perror.stdout

    return 0, process.stdout


#
# get_ovs_definitions()
#
def get_ovs_definitions(objects, pahole="pahole", pid=None):
    if pid is None:
        raise ValueError("A valid pid value should be supplied!")

    if not isinstance(objects, list):
        objects = [objects]

    if len(objects) == 0:
        raise ValueError("Must supply at least one object!")

    vswitchd = Path("/proc/{}/exe".format(str(pid))).resolve()

    object_str = ','.join(objects)

    def run_pahole(debug_file):
        error, result = run_program([pahole, "-C", object_str, "--compile",
                                     debug_file])

        if error:
            if "pahole: {}: Invalid argument".format(debug_file) not in result:
                print("ERROR: Pahole failed to get ovs-vswitchd data "
                      "structures!\n{}".format(re.sub('^', ' ' * 7,
                                                      result.rstrip(),
                                                      flags=re.MULTILINE)))
                sys.exit(-1)

            return None

        if bool(re.search("pahole: type .* not found", result)):
            return None

        return result

    def run_readelf(bin_file):
        error, result = run_program(['readelf', "-n",
                                     "--debug-dump=links", bin_file])

        if error:
            print("ERROR: Failed 'readelf' on \"{}\"!\n{}".
                  format(bin_file, re.sub('^', ' ' * 7, result,
                                          flags=re.MULTILINE)))
            sys.exit(-1)

        return result

    def get_debug_file(bin_file):
        elf_result = run_readelf(bin_file)
        match = re.search("Build ID: ([0-9a-fA-F]+)", elf_result)
        if not match:
            print("ERROR: Can't find build ID to read debug symbols!")
            sys.exit(-1)

        dbg_file = "/usr/lib/debug/.build-id/{}/{}.debug".format(
            match.group(1)[:2], match.group(1)[2:])

        return dbg_file

    def get_from_shared_library(debug_file):
        ovs_libs = ['libofproto', 'libopenvswitch', 'libovsdb', 'libsflow',
                    'libvtep']
        error, ldd_result = run_program(['ldd', debug_file])

        if error:
            print("ERROR: Failed 'ldd' on \"{}\"!\n{}".
                  format(debug_file, re.sub('^', ' ' * 7, ldd_result,
                                            flags=re.MULTILINE)))
            sys.exit(-1)

        for lib in ovs_libs:
            match = re.search(r"^\s*{}.* => (.*) \(.*\)$".format(lib),
                              ldd_result, flags=re.MULTILINE)
            if match is None:
                continue

            result = run_pahole(match.group(1))
            if result is None:
                result = run_pahole(get_debug_file(match.group(1)))

            if result:
                return result

        return None

    #
    # First try to find the debug data as part of the executable.
    #
    result = run_pahole(vswitchd)

    if result is None:
        print("INFO: Failed to find debug info in \"{}\"!".format(vswitchd))

        #
        # Get additional .debug information if available.
        #
        dbg_file = get_debug_file(vswitchd)
        result = run_pahole(dbg_file)
        if result is None:
            print("INFO: Failed to find debug info in \"{}\"!".format(
                dbg_file))

        #
        # Try to get information from shared libraries if used.
        #
        result = get_from_shared_library(vswitchd)

    if result is None:
        print("ERROR: Failed to find needed data structures through pahole!")
        sys.exit(-1)

    #
    # We need an empty _Atomic definition to avoid compiler complaints.
    #
    result = "#define _Atomic\n" + result

    #
    # Remove the uint64_t definition as it conflicts with the kernel one.
    #
    result = re.sub("^typedef.*uint64_t;$", "", result, flags=re.MULTILINE)

    return result


#
# next_power_of_two()
#
def next_power_of_two(val):
    np = 1
    while np < val:
        np *= 2
    return np


#
# dynamic_plot_init()
#
def dynamic_plot_init(real_time=True):

    if real_time:
        lines = []
        fig, axs = plt.subplots(4, figsize=(19, 10))
        fig.suptitle('Revalidator Handling')
        for ax in axs:
            ax.grid()

        axs[0].set_ylabel("Numer of flows", weight='bold')
        axs[1].set_ylabel("Time spend (ms)", weight='bold')
        axs[2].set_ylabel("Numer of flows", weight='bold')
        axs[3].set_ylabel("Time spend (ms)", weight='bold')
        axs[3].set_xlabel("Time (seconds since start)", weight='bold')

        lines.append(axs[0].plot([], [], label="n_flows", marker='o')[0])
        lines.append(axs[0].plot([], [], label="n_reval_flows")[0])
        axs[0].legend(bbox_to_anchor=(1, 1), loc='upper left',
                      borderaxespad=0.5)
        axs[0].set_xlim(0, 30)
        axs[0].set_ylim(-4, 104)

        lines.append(axs[1].plot([], [], color="orange",
                                 label="dump_duration")[0])
        axs[1].legend(bbox_to_anchor=(1, 1),
                      loc='upper left', borderaxespad=0.5)
        axs[1].set_xlim(0, 30)
        axs[1].set_ylim(-0.4, 10.4)

        lines.append(axs[2].plot([], [], label="avg_n_flows")[0])
        lines.append(axs[2].plot([], [], label="max_n_flows")[0])
        lines.append(axs[2].plot([], [], label="flow_limit")[0])
        axs[2].legend(bbox_to_anchor=(1, 1), loc='upper left',
                      borderaxespad=0.5)
        axs[2].set_xlim(0, 30)
        axs[2].set_ylim(-600, 15600)

        lines.append(axs[3].plot([], [], label="poll_wait")[0])
        lines.append(axs[3].plot([], [], label="actual_wait")[0])
        axs[3].legend(bbox_to_anchor=(1, 1), loc='upper left',
                      borderaxespad=0.5)
        axs[3].set_xlim(0, 30)
        axs[3].set_ylim(-20, 520)

        fig.tight_layout()

        plt.ion()
        plt.show()
    else:
        fig = None
        axs = None
        lines = None

    graph_data = {"base_time": 0,
                  "l_index": 0,
                  "fig": fig,
                  "axs": axs,
                  "lines": lines,
                  "last_update": 0,
                  "ukey_count": 0,
                  "time": [],
                  "n_flows": [],
                  "n_reval_flows": [],
                  "avg_n_flows": [],
                  "max_n_flows": [],
                  "flow_limit": [],
                  "dump_duration": [],
                  "poll_wait": [],
                  "actual_wait": []}

    return namedtuple("GraphData", graph_data.keys())(*graph_data.values())


#
# dynamic_plot_update()
#
def dynamic_plot_update(graph_data, refresh=1):

    if graph_data.last_update != 0 and \
       (graph_data.time[-1] - graph_data.last_update) < refresh:
        return graph_data

    graph_data = graph_data._replace(last_update=graph_data.time[-1])

    if (graph_data.time[-1] - graph_data.time[graph_data.l_index]) > 30:
        for i in range(graph_data.l_index + 1, len(graph_data.time)):
            if (graph_data.time[-1] - graph_data.time[i]) <= 30:
                graph_data = graph_data._replace(l_index=i)
                break

    for line in graph_data.lines:
        line.set_xdata(graph_data.time[graph_data.l_index:])

    graph_data.lines[0].set_ydata(graph_data.n_flows[graph_data.l_index:])
    graph_data.lines[1].set_ydata(
        graph_data.n_reval_flows[graph_data.l_index:])
    graph_data.lines[2].set_ydata(
        graph_data.dump_duration[graph_data.l_index:])
    graph_data.lines[3].set_ydata(graph_data.avg_n_flows[graph_data.l_index:])
    graph_data.lines[4].set_ydata(graph_data.max_n_flows[graph_data.l_index:])
    graph_data.lines[5].set_ydata(graph_data.flow_limit[graph_data.l_index:])
    graph_data.lines[6].set_ydata(graph_data.poll_wait[graph_data.l_index:])
    graph_data.lines[7].set_ydata(graph_data.actual_wait[graph_data.l_index:])

    for ax in graph_data.axs:
        if graph_data.l_index == 0:
            ax.autoscale(enable=True, axis='y')
        else:
            ax.autoscale(enable=True)

        ax.relim(visible_only=True)
        ax.autoscale_view(tight=True, scalex=True, scaley=True)

    try:
        graph_data.fig.canvas.draw()
        graph_data.fig.canvas.flush_events()
    except KeyboardInterrupt:
        return None

    return graph_data


#
# show_graph()
#
def show_graph(graph_data, gui=False, file_name=None):

    if len(graph_data.time) == 0 or (not gui and file_name is None):
        return

    plt.ioff()

    fig, (nf_ax, dd_ax, f_ax, t_ax) = plt.subplots(4, figsize=(19, 10))
    fig.suptitle('Revalidator Handling')
    nf_ax.grid()
    f_ax.grid()
    dd_ax.grid()
    t_ax.grid()

    nf_ax.set_ylabel("Numer of flows", weight='bold')
    f_ax.set_ylabel("Numer of flows", weight='bold')
    dd_ax.set_ylabel("Time spend (ms)", weight='bold')
    t_ax.set_ylabel("Time spend (ms)", weight='bold')
    t_ax.set_xlabel("Time (seconds since start)", weight='bold')

    nf_ax.plot(graph_data.time, graph_data.n_flows, label="n_flows")
    nf_ax.plot(graph_data.time, graph_data.n_reval_flows,
               label="n_reval_flows")
    nf_ax.legend(bbox_to_anchor=(1, 1), loc='upper left', borderaxespad=0.5)

    dd_ax.plot(graph_data.time, graph_data.dump_duration, color="orange",
               label="dump_duration")
    dd_ax.legend(bbox_to_anchor=(1, 1), loc='upper left', borderaxespad=0.5)

    f_ax.plot(graph_data.time, graph_data.avg_n_flows, label="avg_n_flows")
    f_ax.plot(graph_data.time, graph_data.max_n_flows, label="max_n_flows")
    f_ax.plot(graph_data.time, graph_data.flow_limit, label="flow_limit")
    f_ax.legend(bbox_to_anchor=(1, 1), loc='upper left', borderaxespad=0.5)

    t_ax.plot(graph_data.time, graph_data.poll_wait, label="poll_wait")
    t_ax.plot(graph_data.time, graph_data.actual_wait, label="actual_wait")
    t_ax.legend(bbox_to_anchor=(1, 1), loc='upper left', borderaxespad=0.5)

    fig.tight_layout()

    if file_name is not None and file_name != "":
        fig.savefig(file_name + '.png')

    if gui:
        try:
            plt.show()
        except KeyboardInterrupt:
            pass

    plt.close(fig)


#
# process_events_from_file()
#
def process_events_from_file(file_name):
    try:
        with open(file_name, 'r') as fd:
            print("- Reading events from \"{}\"...".format(file_name))

            print_csv_header()
            for entry in fd:
                entry.rstrip()
                if entry.startswith('event = {'):
                    event = ast.literal_eval(entry[8:])
                    event = namedtuple("EventObject",
                                       event.keys())(*event.values())
                    try:
                        _process_event(event)
                    except KeyboardInterrupt:
                        break

    except (FileNotFoundError, PermissionError):
        print("ERROR: Can't open file \"{}\" for reading!".format(file_name))
        sys.exit(-1)

    show_graph(graph, gui=not options.no_gui, file_name=options.write_charts)


#
# main()
#
def main():
    #
    # Don't like these globals, but ctx passing does not seem to work with the
    # existing open_ring_buffer() API :(
    #
    global b
    global export_file
    global options
    global state
    global graph

    #
    # Argument parsing
    #
    parser = argparse.ArgumentParser()

    parser.add_argument("-c", "--compress-output", action="store_true",
                        help="Compress output, i.e. only dump changes in "
                        "the dataset")
    parser.add_argument("--buffer-page-count",
                        help="Number of BPF ring buffer pages, default 1024",
                        type=int, default=1024, metavar="NUMBER")
    parser.add_argument("-D", "--debug",
                        help="Enable eBPF debugging",
                        type=int, const=0x3f, default=0, nargs='?')
    parser.add_argument("-g", "--no-gui", action="store_true",
                        help="Do not use the gui to display plots")
    parser.add_argument("--no-ukey-count", action="store_true",
                        help="No revalidate_ukey() counting")
    parser.add_argument("-p", "--pid", metavar="VSWITCHD_PID",
                        help="ovs-vswitch's PID",
                        type=int, default=None)
    parser.add_argument("-P", "--pahole", metavar="PAHOLE",
                        help="Pahole executable to use, default pahole",
                        type=str, default="pahole")
    parser.add_argument("-r", "--read-events",
                        help="Read events from <FILE> instead of installing "
                        "tracepoints", type=str, default=None, metavar="FILE")
    parser.add_argument("-R", "--no-realtime-plots", action="store_true",
                        help="Do not show realtime plot while tracing")
    parser.add_argument("-u", "--update-interval",
                        help="Seconds to wait between real time update, "
                        "default 1", type=float, default=1, metavar="SECONDS")
    parser.add_argument("-w", "--write-events",
                        help="Write events to <FILE>",
                        type=str, default=None, metavar="FILE")
    parser.add_argument("-W", "--write-charts",
                        help="Write overall charts to <FILE>.png",
                        type=str, default=None, metavar="FILE")

    options = parser.parse_args()

    #
    # Find the PID of the ovs-vswitchd daemon if not specified.
    #
    if options.pid is None and options.read_events is None:
        for proc in psutil.process_iter():
            if 'ovs-vswitchd' in proc.name():
                if options.pid is not None:
                    print("ERROR: Multiple ovs-vswitchd daemons running, "
                          "use the -p option!")
                    sys.exit(-1)

                options.pid = proc.pid

    #
    # Error checking on input parameters.
    #
    if options.pid is None and options.read_events is None:
        print("ERROR: Failed to find ovs-vswitchd's PID!")
        sys.exit(-1)

    if options.read_events is not None and options.write_events is not None:
        print("ERROR: Either supply the read or write events option, "
              "not both!")
        sys.exit(-1)

    options.buffer_page_count = next_power_of_two(options.buffer_page_count)

    #
    # Define the state and graph.
    #
    state = {"last_start": None,
             "last_done": None,
             "running": False,
             "last_csv": None,
             "last_not_printed_csv": None}

    export_file = None

    graph = dynamic_plot_init(real_time=(not options.no_gui
                                         and not options.no_realtime_plots))

    #
    # Process events from file if required.
    #
    if options.read_events is not None:
        process_events_from_file(options.read_events)
        sys.exit(0)

    #
    # Open write handle if needed.
    #
    if options.write_events is not None:
        try:
            export_file = open(options.write_events, "w")
        except (FileNotFoundError, IOError, PermissionError) as e:
            print("ERROR: Can't create export file \"{}\": {}".format(
                options.write_events, e.strerror))
            sys.exit(-1)

    #
    # Attach the usdt probe.
    #
    u = USDT(pid=int(options.pid))
    try:
        u.enable_probe(probe="start_dump", fn_name="probe__start_dump")
        u.enable_probe(probe="sweep_done", fn_name="probe__sweep_done")
        if not options.no_ukey_count:
            u.enable_probe(probe="revalidate_ukey__:entry",
                           fn_name="probe__reval_entry")
    except USDTException as e:
        print("ERROR: {}".format(
            (re.sub('^', ' ' * 7, str(e), flags=re.MULTILINE)).strip().
            replace("--with-dtrace or --enable-dtrace",
                    "--enable-usdt-probes")))
        sys.exit(-1)

    #
    # Attach probe to running process.
    #
    source = EBPF_SOURCE.replace("<EVENT_ENUM>", "\n".join(
        ["    EVENT_{} = {},".format(
            event.name, event.value) for event in Event]))
    source = source.replace("<BUFFER_PAGE_CNT>",
                            str(options.buffer_page_count))
    source = source.replace("<OVS_INCLUDE_DEFINITIONS>",
                            get_ovs_definitions("udpif", pid=options.pid,
                                                pahole=options.pahole))

    b = BPF(text=source, usdt_contexts=[u], debug=options.debug)

    #
    # Print header.
    #
    ltz = datetime.datetime.now()
    utc = ltz.astimezone(pytz.utc)
    time_string = "# Starting trace @{} ({} UTC)".format(
        ltz.isoformat(), utc.strftime("%H:%M:%S"))

    if export_file is not None:
        export_file.write(time_string + "\n")

    print(time_string)
    print_csv_header()

    #
    # Process all events.
    b['events'].open_ring_buffer(process_event)
    while 1:
        try:
            b.ring_buffer_poll()
        except KeyboardInterrupt:
            break

    dropcnt = b.get_table("dropcnt")
    for k in dropcnt.keys():
        count = dropcnt.sum(k).value
        if k.value == 0 and count > 0:
            print("\n# WARNING: Not all upcalls were captured, {} were "
                  "dropped!\n#          Increase the BPF ring buffer size "
                  "with the --buffer-page-count option.".format(count))

    #
    # Display footer.
    #
    if state["last_not_printed_csv"] is not None:
        print(state["last_not_printed_csv"])

    ltz = datetime.datetime.now()
    utc = ltz.astimezone(pytz.utc)
    time_string = "# Stopping trace @{} ({} UTC)".format(
        ltz.isoformat(), utc.strftime("%H:%M:%S"))

    if export_file is not None:
        export_file.write(time_string + "\n")

    print(time_string)

    #
    # Close event file is used.
    #
    if options.write_events is not None:
        export_file.close()

    #
    # Do final graph if requested.
    #
    show_graph(graph, gui=not options.no_gui, file_name=options.write_charts)


#
# Start main() as the default entry point...
#
if __name__ == '__main__':
    main()



# Copyright (c) 2009, 2010 Nicira, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at:
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import getopt
import sys

import ovs.json


def print_json(json):
    if isinstance(json, str):
        print("error: %s" % json)
        return False
    else:
        ovs.json.to_stream(json, sys.stdout)
        sys.stdout.write("\n")
        return True


def parse_multiple(stream):
    buf = stream.read(4096)
    ok = True
    parser = None
    while len(buf):
        if parser is None and buf[0] in " \t\r\n":
            buf = buf[1:]
        else:
            if parser is None:
                parser = ovs.json.Parser()
            n = parser.feed(buf)
            buf = buf[n:]
            if len(buf):
                if not print_json(parser.finish()):
                    ok = False
                parser = None
        if len(buf) == 0:
            buf = stream.read(4096)
    if parser and not print_json(parser.finish()):
        ok = False
    return ok


def main(argv):
    argv0 = argv[0]

    try:
        options, args = getopt.gnu_getopt(argv[1:], '', ['multiple'])
    except getopt.GetoptError as geo:
        sys.stderr.write("%s: %s\n" % (argv0, geo.msg))
        sys.exit(1)

    multiple = False
    for key, value in options:
        if key == '--multiple':
            multiple = True
        else:
            sys.stderr.write("%s: unhandled option %s\n" % (argv0, key))
            sys.exit(1)

    if len(args) != 1:
        sys.stderr.write("usage: %s [--multiple] INPUT.json\n" % argv0)
        sys.exit(1)

    input_file = args[0]
    if input_file == "-":
        stream = sys.stdin
    else:
        stream = open(input_file, "r")

    if multiple:
        ok = parse_multiple(stream)
    else:
        ok = print_json(ovs.json.from_stream(stream))

    if not ok:
        sys.exit(1)


if __name__ == '__main__':
    main(sys.argv)



#!/usr/bin/env python3
# Copyright (c) 2024 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at:
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from pathlib import Path
import os
import sys

# The tester might want to select a PCI device, if so, trust it.
if 'DPDK_PCI_ADDR' in os.environ:
    print(os.environ['DPDK_PCI_ADDR'])
    sys.exit(0)

for device in sorted(Path('/sys/bus/pci/devices').iterdir()):
    class_path = device / 'class'
    # Only consider Network class devices
    if class_path.read_text().strip() != '0x020000':
        continue
    kmod_path = device / 'driver' / 'module'
    kmod_name = kmod_path.resolve().name
    # Only care about devices bound to vfio_pci or igb_uio.
    if kmod_name not in ['vfio_pci', 'igb_uio']:
        continue
    print(device.resolve().name)
    sys.exit(0)

sys.exit(1)
